PAGE	58,91
.286
TITLE	COPY-PROTECTED TRACK FORMAT

; Make a copy-protected track on a 720 KB disk.
; Format a 512 B sector at 300 kHz, then another
; one at 250 kHz, and an 8 KB sector.  Dump 512 B
; of garbage and some codes to the last one.  See
; the code for details: there are plenty of them!
;
; NOTE: This method is not mine.  Microcosm Ltd.,
; 17 Cranbrook Road, Bristol, England, have used
; it in their CopyControl 1.61 routine.  Thanks!
;
; Requirements: AT BIOS dated after 15.11.1986,
; Microsoft (R) Macro Assembler Version 5.10 or ^
; Copyright (C) MXMV Luchezar Georgiev, Bulgaria
;________________e-mail: <lucho@midi.tu-varna.bg>
;_________________________________________________
; IMPORTANT: Define TRACK and SIDE in the Makefile

DUMPCHR EQU	'Z'	; character to dump
DUMPSIZ EQU	512-1	; how much of dump before
			; the special code area
SECTORS EQU	3
LONGLEN EQU	6	; long sector is 8 KB

SECTIME EQU	16	; between 2 sectors [ms]

FDCBASE EQU	3F0H
FDCSTAT EQU	FDCBASE+4	; the two
FDCDATA EQU	FDCBASE+5	; 8272A ports
FDCRATE EQU	FDCBASE+7	; rate control reg.

SPECIFY EQU	00000011B	; commands
FMTRACK EQU	01001101B
WRTDATA EQU	01000101B

DATADIR EQU	01000000B	; status bit masks
DATARDY EQU	10000000B

SUCCESS EQU	11000000B	; status reg.0 code
BADCRC	EQU	10H		; int 13 error code

RATE250 EQU	10B		; 250 kHz rate
RATE300 EQU	01B		; 300 kHz rate
RATE500 EQU	00B		; 500 kHz rate

CODE	SEGMENT
ASSUME	CS:CODE,DS:CODE

ORG	100H			; it's a .COM file
PUBLIC	START			; MSYM needs this

START:	CLD			; increment indexes
	LEA	DI,OLD1EH
	LDS	SI,INT1EV
	MOVSW			; save old DPT
	MOVSW			; address
	PUSH	CS
	POP	DS		; restore DS
	LES	DI,INT1EV
	LEA	AX,DPT		; place our DPT
	STOSW			; offset
	MOV	AX,CS		; and segment
	STOSW
	MOV	AX,DS:[82H]	; drive from PSP
	AND	AL,NOT('a'-'A') ; make it uppercase
	SUB	AX,(':' SHL 8)+'A' ; is it drive 0?
	JZ	DRVOK		; yes, don't change
	CMP	AL,1		; no, drive 1?
	JNZ	@F		; no, abort
	MOV	DRIVE,AL	; yes, update drive
	OR	FPDRIVE,AL	; in all the places
	OR	WPDRIVE,AL
DRVOK:	MOV	CH,79		; max track #
	MOV	CL,9		; max sector #
	MOV	DL,DRIVE
	MOV	AH,18H		; get DPT in ES:DI
	CALL	INT40H		; can we use 720K?
	JC	@F		; no, sorry...
	PUSH	CS		; discard BIOS's
	POP	ES		; info, restore ES
	CALL	GETPARM
	MOV	AX,(5 SHL 8)+SECTORS
	CALL	INT40H		; format normally
	JC	@F
	CALL	FORMAT		; format the copy-
	JC	@F		; protected track
	MOV	AL,LONGLEN
	MOV	SECTLEN,AL	; 8 KB
	CALL	WRITE8K		; write 8 KB to
	CALL	GETPARM		; 3rd sector (#9)
	MOV	AX,1		; (ignore errors)
	CALL	INT40H		; reset FDC first
	JC	@F
	MOV	CL,9
	MOV	AH,4		; verify the 8 KB
	CALL	INT40H		; sector
	CMP	AH,BADCRC	; only bad CRC?
	JNE	@F		; not only... :-(
	MOV	AL,2		; yes, switch
	MOV	SECTLEN,AL	; back to 512 B
	MOV	CL,2
	MOV	AX,401H
	CALL	INT40H		; verify 2nd sector
	JC	@F
	MOV	AL,RATE300
	CALL	NEWRATE		; switch to 300 kHz
	DEC	CL
	MOV	AX,401H
	CALL	INT40H		; verify 1st sector
	JNC	FINISH		; ok?
@@:	MOV	AH,0		; no, reset disk
	CALL	INT40H
	MOV	AL,7		; undocumented int!
	INT	29H		; beep (ASCII BEL)
FINISH: LES	DI,INT1EV
	LEA	SI,OLD1EH
	MOVSW			; restore old DPT
	MOVSW
	MOV	AL,RATE250	; restore to 250
	CALL	NEWRATE		; kHz
	PUSH	0
	RET			; terminate: int 20

DRIVE	DB	0

INT1EV	DD	4*1EH		; int 1Eh vector
OLD1EH	DD	?		; old int 1Eh
				; handler
;
; Our disk parameter table
;
DPT	DB	0DFH,2,25H
SECTLEN	DB	2		; to be changed
	DB	SECTORS,2,0FFH,28H,2,0FH,8

;
; This buffer is required by the FDC during format
;
BUF	DB		TRACK,SIDE,1,2	; 512 B 
	DB		TRACK,SIDE,2,2	; 512 B
	DB		TRACK,SIDE,9,LONGLEN ; 8K
;				   ^sector number
;
;
; Format a track directly interfacing with the FDC.
; A 512 B sector at 300 kHz, then another 2 at 250.
; No DMA used, no Int 0EH used -- just polling!
;
FMTPREP DB	SPECIFY,0DFH,3,FMTRACK
FPDRIVE	DB	SIDE SHL 2,2,SECTORS,28H
FPREND	DB	2

FORMAT	PROC	NEAR
	LEA	SI,FMTPREP
@@:	CALL	FDCOUT		; prepare the FDC
	CMP	SI,OFFSET FPREND
	JB	@B
	CLI			; disable interrupt
	CALL	FDCOUT		; start execution
	MOV	AL,RATE300
	CALL	NEWRATE		; switch to 300 kHz
	LEA	SI,BUF
FMTNXT: CMP	SI,OFFSET BUF+4 ; is sector 2 next?
	JNZ	@F		; no
	XOR	CX,CX		; yes
	MOV	DX,SECTIME*1000
	MOV	AH,86H
	INT	15H		; let sector #1 end
	CLI			; re-disable ints
	MOV	AL,RATE250
	CALL	NEWRATE		; back to 250 kHz
@@:	CALL	FDCOUT		; send parameter
	CMP	SI,OFFSET BUF+4*SECTORS ; done?
	JB	FMTNXT		; not yet
RESULT:	STI			; yes, allow ints
	CALL	FDCIN		; get result
	MOV	AH,AL		; save it
	MOV	BX,6		; 6 more bytes left
@@:	CALL	FDCIN		; - discard them
	DEC	BX
	JNZ	@B
	AND	AH,SUCCESS	; OK?
	JZ	FMTRET		; yes
	STC			; no
FMTRET:	RET
FORMAT	ENDP

;
; Write an 8 KB long sector, following the format.
; After writing some garbage and codes necessary,
; double the speed so that the operation completes
; before the index hole comes for the second time.
;
WRTPREP DB	WRTDATA
WPDRIVE DB	SIDE SHL 2,TRACK,SIDE,9,LONGLEN,9
	DB	2,0FFH
WRPREND	DB	DUMPCHR

WRITE8K PROC	NEAR
	LEA	SI,WRTPREP
@@:	CALL	FDCOUT		; prepare the FDC
	CMP	SI,OFFSET WRPREND
	JB	@B
	MOV	BX,2 SHL(LONGLEN+6) ; load counter
	CLI			; disable interrupt
WRTNXT:	CALL	FDCOUT		; dump it all
	DEC	SI		; back to dump char
	CMP	BX,2 SHL(LONGLEN+6)-DUMPSIZ ;e-o-d?
	JNE	WCDOWN		; no
	LEA	SI,CODAREA	; yes, write
@@:	CALL	FDCOUT		; special codes
	DEC	BX
	CMP	SI,OFFSET CODEND ; codes written?
	JB	@B		; not yet
	MOV	AL,RATE500	; yes, double speed
	CALL	NEWRATE		; (rest is garbage)
WCDOWN:	DEC	BX		; count down; done?
	JNZ	WRTNXT		; not yet
	MOV	AL,RATE250	; yes, switch back
	CALL	NEWRATE		; to normal rate
	JMP	RESULT		; yes, get results
WRITE8K ENDP

;
; Send value pointed to by DS:SI to the FDC.
; Timeout if never ready, returning 2 levels up,
; to the main routine, skiping the caller
;
OUTCNTR DB	50		; counter hi byte

FDCOUT	PROC	NEAR
	MOV	DX,FDCSTAT
OUTTRY:	XOR	CX,CX		; counter low byte
OUTCHK:	IN	AL,DX		; get status
	TEST	AL,DATADIR	; direction ok?
	JNZ	@F		; no
	TEST	AL,DATARDY	; yes, ready?
	JNZ	OK2OUT		; yes, go output
@@:	LOOP	OUTCHK		; no, try again
	DEC	OUTCNTR		; more to wait?
	JNS	OUTTRY		; yes
	STC			; no, timeout
	STI			; might be disabled
	POP	AX		; discard return
	RET			; address
OK2OUT: MOV	DX,FDCDATA
	OUTSB			; do it; C=0 from
	RET			; test instruction
FDCOUT	ENDP

;
; Get one byte from the FDC.  Upon timeout, return
; two levels up (to the main routine)
;
INCNTR	DB	50		; counter hi byte

FDCIN	PROC	NEAR
	MOV	DX,FDCSTAT
INTRY:	XOR	CX,CX
@@:	IN	AL,DX
	AND	AL,DATARDY+DATADIR ; ready with
	CMP	AL,DATARDY+DATADIR ; next byte?
	JZ	@F		; yes
	LOOP	@B		; no, try again
	DEC	INCNTR		; more to wait?
	JNS	INTRY		; yes
	STC			; no, timeout
	POP	AX		; return 2 levels
	RET			; up
@@:	MOV	DX,FDCDATA
	IN	AL,DX		; get data byte
	RET
FDCIN	ENDP

;
; Set new FDC transfer rate (250 or 300 kHz)
;
NEWRATE PROC	NEAR
	PUSH	DX
	MOV	DX,FDCRATE
	OUT	DX,AL
	POP	DX
	RET
NEWRATE ENDP

;
; Get parameters required by Int 40H.
; Only AX remains uninitialized.
;
GETPARM PROC	NEAR
	LEA	BX,BUF
	MOV	CX,(TRACK SHL 8)+1
	MOV	DL,DRIVE
	MOV	DH,SIDE
	RET
GETPARM ENDP

;
; Call Int 40H service (its entry
; point appears to be constant)
;
INT40H	PROC	NEAR
	PUSHF
	PUSH	CS
	PUSH	OFFSET @F
	PUSH	0F000H
	PUSH	0EC59H
	RETF
@@:	RET
INT40H	ENDP	

;
; The special code area
;
CODAREA DB	'NNNN',3 DUP (0,0,'NNN')
	DD	0
CODEND	DB	0

CODE	ENDS
	END	START
